//+------------------------------------------------------------------------
//
//  Copyright @ Baris TAZE
//
//  File:     Population.h
//
//  Content:  Interface and Implementation for CPopulation class
//  
//  Classes:  CPopulation
//            
//  History:  27-May-2002    created by baris taze .  
//
//-------------------------------------------------------------------------

#if !defined(__POPULATION_H__)
#define      __POPULATION_H__

#define RANDOM_INT_PRECISION_ER		10000000			

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

///////////////////////////////////
template<class TINDIV, class TBIT >
///////////////////////////////////
class CPopulation  
{
public:
	CPopulation();
	~CPopulation();
	
	void Free();

	void CreateIndividuals( int nIndividualCount );
	
	double CalculateBias( );				
	double CalculateAvgPerf( ); 
	
	void SetIndividual( int nIndex, TINDIV* pPerson );
	TINDIV* GetIndividual( int nIndex );

	int GetIndividualCount();
	void Lock( int nBitIndex, bool bPerfect);
	void Unlock( int nBitIndex );
	
	void Sort();

	CString ToString();

	//NaturalSelection Algorithms...
	void RouletteWheelProportionate( double fAveragePerformance );
	void StochasticRemainderSelection( double fAveragePerformance );
	void GenitorSelection( int nReplaceCount );
	void TournamentSelection( double fChanceOfGood );
	
	TINDIV* GetRandomIndividual();

private:
	static int	RoundDoubleToInt( double fNumber );
	TINDIV**	m_pIndividuals;
	int			m_nIndividualCount;
};


///////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
// Function Name:	Constructor
// Parameters	:	void
// Return Type	:	void
// Description	:	Constructor	
// Assumptions	:	No Assumption
///////////////////////////////////
template<class TINDIV, class TBIT >
///////////////////////////////////
CPopulation<TINDIV, TBIT>::CPopulation()
{	
	m_pIndividuals     = NULL;
	m_nIndividualCount = 0;
}

//////////////////////////////////////////////////////////////////////
// Function Name:	RoundDoubleToInt
// Parameters	:	double, 6.97f
// Return Type	:	int, 7 
// Description	:	rounds the double number to the int
// Assumptions	:	No Assumption
///////////////////////////////////
template<class TINDIV, class TBIT >
///////////////////////////////////
int CPopulation<TINDIV, TBIT>::RoundDoubleToInt( double fNumber )
{
	return ( int )(fNumber + 0.5f);
}

//////////////////////////////////////////////////////////////////////
// Function Name:	CreateIndividuals
// Parameters	:	int nIndividualCount
// Return Type	:	void
// Description	:	creates the all individuals' pointers
// Assumptions	:	No Assumption
///////////////////////////////////
template<class TINDIV, class TBIT >
///////////////////////////////////
void CPopulation<TINDIV, TBIT>::CreateIndividuals( int nIndividualCount )
{
	ASSERT( nIndividualCount > 0 );
	//fatal error
	
	ASSERT( NULL == m_pIndividuals );
	//call this function only once.

	m_nIndividualCount = nIndividualCount;
	m_pIndividuals = new TINDIV*[ m_nIndividualCount ];
	for( int x=0; x<m_nIndividualCount; x++ )
	{
		m_pIndividuals[ x ] = NULL;
	}
}

//////////////////////////////////////////////////////////////////////
// Function Name:	Destructor
// Parameters	:	void
// Return Type	:	void
// Description	:	Destructor
// Assumptions	:	No Assumption
///////////////////////////////////
template<class TINDIV, class TBIT >
///////////////////////////////////
CPopulation<TINDIV, TBIT>::~CPopulation()
{
	Free();
}

///////////////////////////////////
template<class TINDIV, class TBIT >
///////////////////////////////////
void CPopulation<TINDIV, TBIT>::Free()
{
	if( NULL != m_pIndividuals )
	{
		for( int x=0; x<m_nIndividualCount; x++ )
		{
			if( NULL != m_pIndividuals[ x ] )
			{
				delete m_pIndividuals[ x ];
				m_pIndividuals[ x ] = NULL;
			}			
		}

		delete[] m_pIndividuals;
		m_pIndividuals = NULL;
	}

	m_nIndividualCount = 0;
}

//////////////////////////////////////////////////////////////////////
// Function Name:	SetIndividual
// Parameters	:	int nIndex, position to be set.
//					TINDIV* pPerson, the object to be set.
// Return Type	:	void
// Description	:	sets the individual to the desired location.
// Assumptions	:	Assumed that, "CreateIndividuals()" has been already called. 
//					Assumed that, the old object is freed explicitly or 
//					it is still on use. [ the old one have to be explicitly deleted ]
//					It may be NULL, too.
///////////////////////////////////
template<class TINDIV, class TBIT >
///////////////////////////////////
void CPopulation<TINDIV, TBIT>::SetIndividual( int nIndex, TINDIV* pPerson )
{
	ASSERT( nIndex >= 0 );
	//fatal error

	ASSERT( nIndex < m_nIndividualCount );
	//fatal error

	ASSERT( NULL != m_pIndividuals );
	//make sure that CreateIndividuals() has been already called.

	m_pIndividuals[ nIndex ] = pPerson;
}

//////////////////////////////////////////////////////////////////////
// Function Name:	GetIndividual
// Parameters	:	int nIndex, position of the individual
// Return Type	:	TINDIV* //may be NULL;
// Description	:	gets the individual from the desired location.
// Assumptions	:	Assumed that, "CreateIndividuals()" has been already called. 
///////////////////////////////////
template<class TINDIV, class TBIT >
///////////////////////////////////
TINDIV* CPopulation<TINDIV, TBIT>::GetIndividual( int nIndex )
{
	ASSERT( nIndex >= 0 );
	//fatal error

	ASSERT( nIndex < m_nIndividualCount );
	//fatal error

	ASSERT( NULL != m_pIndividuals );
	//make sure that CreateIndividuals() has been already called.

	return m_pIndividuals[ nIndex ];
}

//////////////////////////////////////////////////////////////////////
// Function Name:	GetIndividualCount
// Parameters	:	void
// Return Type	:	int
// Description	:	gets the individual's count
// Assumptions	:	No Assumption
///////////////////////////////////
template<class TINDIV, class TBIT >
///////////////////////////////////
int CPopulation<TINDIV, TBIT>::GetIndividualCount()
{
	return m_nIndividualCount;
}

//////////////////////////////////////////////////////////////////////
// Function Name:	Lock
// Parameters	:	int nBitIndex
//					bool bPerfect: the perfect bit, which have to be set all individuals.
// Return Type	:	void
// Description	:	Sets the INDEXth bit of the all individuals with "bPerfect" and,
//					makes their state as "locked".
// Assumptions	:	Assumed that, "CreateIndividuals()" has been already called. 
//					Assumed that, all individuals are created explicitly and set.	
///////////////////////////////////
template<class TINDIV, class TBIT >
///////////////////////////////////
void CPopulation<TINDIV, TBIT>::Lock( int nBitIndex, bool bPerfect )
{
	ASSERT( nBitIndex >= 0 );
	//fatal error

	ASSERT( NULL != m_pIndividuals );
	//make sure that CreateIndividuals() has been already called.

	for( int x=0; x<m_nIndividualCount; x++ )
	{
		TBIT* pTemp = m_pIndividuals[ x ]->GetBit( nBitIndex );
		pTemp->SetValue( bPerfect );
		pTemp->Lock();		
	}
}

//////////////////////////////////////////////////////////////////////
// Function Name:	UnLock
// Parameters	:	int nBitIndex
// Return Type	:	void
// Description	:	unlock INDEXth bit of the all individuals.
// Assumptions	:	Assumed that, "CreateIndividuals()" has been already called. 
//					Assumed that, all individuals are created explicitly and set.	
///////////////////////////////////
template<class TINDIV, class TBIT >
///////////////////////////////////
void CPopulation<TINDIV, TBIT>::Unlock( int nBitIndex )
{
	ASSERT( nBitIndex >= 0 );
	//fatal error

	ASSERT( NULL != m_pIndividuals );
	//make sure that CreateIndividuals() has been already called.

	for( int x=0; x<m_nIndividualCount; x++ )
	{
		m_pIndividuals[ x ]->UnLock( nBitIndex, 1 );		
	}
}

//////////////////////////////////////////////////////////////////////
// Function Name:	ToString
// Parameters	:	void
// Return Type	:	CString
// Description	:	Returns the bits like 
//					"110110"
//					"010101"
//					"101010"
// Assumptions	:	No Assumption
///////////////////////////////////
template<class TINDIV, class TBIT >
///////////////////////////////////
CString CPopulation<TINDIV, TBIT>::ToString()
{
	CString temp;

	for( int x=0; x<m_nIndividualCount; x++ )
	{
		temp += m_pIndividuals[ x ]->ToString();
		temp += "\r\n";
	}

	return temp;
}

//////////////////////////////////////////////////////////////////////
// Function Name:	CalculateBias
// Parameters	:	void
// Return Type	:	double
// Description	:	
//					1.String		 : "  1    1    1    0    1   0  "
//					2.String		 : "  0    1    0    0    1   0  "
//					3.String		 : "  1    1    1    0    1   0  "
//					4.String		 : "  1    1    0    1    1   0  "
//					--------------------------------------------------
//                  column similarity:  0.75  1.0  0.5  0.75 1.0 1.0
//					--------------------------------------------------
//					similarity(bias) : ( 0.75 + 1.0 + 0.5 + 0.75 +1.0 + 1.0 ) / 6 = 0.8333
//					Am I wrong?
//
// Assumptions	:	Assumed that, all individuals are created explicitly and set.
///////////////////////////////////
template<class TINDIV, class TBIT >
///////////////////////////////////
double CPopulation<TINDIV, TBIT>::CalculateBias( )			 
{
	double fBias = 0.0f;	

	if( 0 == m_nIndividualCount )
	{
		return fBias;
	}
	
	int nBitSize = GetIndividual( 0 )->GetBitSize();
	double* pBitSum = new double[ nBitSize ];

	for( int i=0; i<nBitSize; i++ )
	{
		pBitSum[ i ] = 0.0f;	
	}

	for( int x=0; x<m_nIndividualCount; x++ )
	{
		TINDIV* pPerson = GetIndividual( x );
		for( int y=0; y<nBitSize; y++ )
		{
			TBIT* pBit = pPerson->GetBit( y );
			if( pBit->GetValue() )
			{
				pBitSum[ y ] += 1.0f ;
			}
		}
	}
	
	for( int i=0; i<nBitSize; i++ )
	{
		pBitSum[ i ] /= (double)m_nIndividualCount;	
		if( 0.5f > pBitSum[ i ] )
		{
			pBitSum[ i ] = 1.0f - pBitSum[ i ];
		}

		//iterative averaging...
		fBias = ( ( fBias * i ) + pBitSum[ i ] ) / ( i + 1 );
	}
	
	delete[] pBitSum;
	return fBias;

}

//////////////////////////////////////////////////////////////////////
// Function Name:	CalculateAvgPerf
// Parameters	:	void
// Return Type	:	double
// Description	:	Average the performances
// Assumptions	:	Assumed that, "CreateIndividuals()" has been already called. 
//					Assumed that, all individuals are created explicitly and set.	
///////////////////////////////////
template<class TINDIV, class TBIT >
///////////////////////////////////
double CPopulation<TINDIV, TBIT>::CalculateAvgPerf( )
{
	double fPerformance = 0.0f;
	for( int x=0; x<m_nIndividualCount; x++ )
	{
		TINDIV* pPerson = GetIndividual( x );
		double fPerfX = pPerson->GetPerformance( );
		//iterative averaging...
		fPerformance = ( ( fPerformance * x ) + fPerfX ) / ( x + 1.0f );		
	}

	return fPerformance;
}

//////////////////////////////////////////////////////////////////////
// Function Name:	Sort [ in descending order ]
// Parameters	:	void
// Return Type	:	void
// Description	:	sort according to the performances..
// Assumptions	:	Assumed that, all individuals are created explicitly and set.	
///////////////////////////////////
template<class TINDIV, class TBIT >
///////////////////////////////////
void CPopulation<TINDIV, TBIT>::Sort()
{
	for( int x=0; x<m_nIndividualCount; x++ )
	{
		TINDIV* pFirst = GetIndividual( x );
		for( int y=x+1; y<m_nIndividualCount; y++ )
		{
			TINDIV* pSecond = GetIndividual( y );
			if( pFirst->GetPerformance() < pSecond->GetPerformance() )//swap
			{				
				SetIndividual( x, pSecond );
				SetIndividual( y, pFirst );
				pFirst = pSecond;
			}
		}
	}
}

//////////////////////////////////////////////////////////////////////
// Function Name:	StochasticRemainderSelection
// Parameters	:	double fAveragePerformance, of the population
// Return Type	:	void
// Description	:	a natural selection method
//					---------------------------
//					Example:
//					---------------------------
//					6/10 of individuals are same, i.e. type A, has performance 12.
//					other 4/10  of individuals are same too, i.e. type B, has performance 7.
//					weighted performanse of A = 6x12/100 = 0.72
//					weighted performanse of B = 4x7 /100 = 0.28
//					---------------------------
//					So:
//						The next generation should consist of 7/10 A (= round 7.2 to 7) ,and
//															  2/10 B (= round 2.8 to 2).
//															[ 9/10 ] ???		
//						The remaining 1 individual should be either type A, with the probability 0.2 ( = 7.2 - 7.0 )
//																 or type B, with the probability 0.8 ( = 2.8 - 2.0 )
//						Toss an UNFAIR coin (0.2/0.8 weighted) to handle this odds in average.
//					---------------------------
//					The final distribution will be either 7-3 or 8-2
//					---------------------------
//
// Assumptions	:	Assumed that, all individuals are created explicitly and set.	
//					Assumed that, all individual's performances have already calculated.
///////////////////////////////////
template<class TINDIV, class TBIT >
///////////////////////////////////
void CPopulation<TINDIV, TBIT>::StochasticRemainderSelection( double fAveragePerformance )
{
	double sum		= fAveragePerformance * m_nIndividualCount;
	double weight	= 0.0f;
	int   nCopy		= 0;
	int	  nNewTotal = 0;
	
	CPtrList list;
	TINDIV** pTempPop = new TINDIV*[ m_nIndividualCount ];
	
	int x=0;
	for( x=0; x<m_nIndividualCount; x++ )
	{
		TINDIV* pPerson = m_pIndividuals[ x ];		
		if( !pPerson->IsChecked() )
		{
			nCopy = 1;		
			pPerson->SetChecked( true );			
			weight = pPerson->GetPerformance() / sum;
			for( int y=x+1; y<m_nIndividualCount; y++ )
			{
				TINDIV* pTemp = m_pIndividuals[ y ];
				if( !pTemp->IsChecked() )
				{
					if( pPerson->IsEqual( pTemp ) )
					{
						pTemp->SetChecked( true );
						nCopy++;
					}
				}			
			}//for				
			weight *= (double)nCopy; //the percentage of existance in the next generation
			double weightEx = weight * (double)m_nIndividualCount;			
			int nNewCopy = (int)weightEx;
			int nRemainder = (int)( ( weightEx - (double)nNewCopy ) * 10.0f );
			int z=0;
			for( z=0; z<nNewCopy; z++ )
			{
				pTempPop[ nNewTotal++ ] = pPerson->Clone();									
			}
			for( z=0; z<nRemainder; z++ )
			{
				list.AddTail( pPerson );				
			}			
		}//if !checked		
	}	
		
	for( x=nNewTotal; x<m_nIndividualCount; x++ )
	{	
		int nSelection = rand() % list.GetCount();
		POSITION pos = list.FindIndex( nSelection );
		TINDIV* pSelection = ( TINDIV* )list.GetAt( pos );
		pTempPop[ x ] = pSelection->Clone();
	}

	for( x=0; x<m_nIndividualCount; x++  )
	{
		TINDIV* pTemp = m_pIndividuals[ x ];
		delete pTemp;
		m_pIndividuals[ x ] = pTempPop[ x ]; //SetIndividual
		m_pIndividuals[ x ]->SetChecked( false );
	}
	delete[] pTempPop;
}

//////////////////////////////////////////////////////////////////////
// Function Name:	RouletteWheelProportionate
// Parameters	:	double fAveragePerformance, of the population
// Return Type	:	void
// Description	:	a natural selection method.
//					---------------------------
//					Example:
//					---------------------------
//					6/10 of individuals are same, i.e. type A, has performance 12.
//					other 4/10  of individuals are same too, i.e. type B, has performance 7.
//					weighted performanse of A = 6x12/100 = 0.72
//					weighted performanse of B = 4x7 /100 = 0.28
//					---------------------------
//					So:
//						The next generation should consist of 7/10 A, 3/10 B, in average.
//						Toss an UNFAIR coin (0.72/0.28 weighted) to handle this odds in average.
//					---------------------------
//					However:
//						Since we still coin a toss, it may be 0/10 A and 10/0 B, and vice versa.
//					---------------------------
// Assumptions	:	Assumed that, all individuals are created explicitly and set.	
//					Assumed that, all individual's performances have already calculated.
///////////////////////////////////
template<class TINDIV, class TBIT >
///////////////////////////////////
void CPopulation<TINDIV, TBIT>::RouletteWheelProportionate( double fAveragePerformance )
{
	double sum		= fAveragePerformance * m_nIndividualCount;
	double weight	= 0.0f;
	int   nCopy		= 0;
	int	  nNewTotal = 0;
	
	CPtrList list;

	int x=0;
	for( x=0; x<m_nIndividualCount; x++ )
	{
		TINDIV* pPerson = m_pIndividuals[ x ];		
		if( !pPerson->IsChecked() )
		{
			nCopy = 1;		
			pPerson->SetChecked( true );			
			weight = pPerson->GetPerformance() / sum;
			for( int y=x+1; y<m_nIndividualCount; y++ )
			{
				TINDIV* pTemp = m_pIndividuals[ y ];
				if( !pTemp->IsChecked() )
				{
					if( pPerson->IsEqual( pTemp ) )
					{
						pTemp->SetChecked( true );
						nCopy++;
					}
				}			
			}//for	
			weight *= (double)nCopy; //the percentage of existance in the next generation
			int nNewCopy = CPopulation<TINDIV, TBIT>::RoundDoubleToInt( weight * (double)m_nIndividualCount );			
			for( int z=0; z<nNewCopy; z++ )
			{
				list.AddTail( pPerson );
			}
			
		}//if !checked		
	}	
	
	TINDIV** pTempPop = new TINDIV*[ m_nIndividualCount ];

	for( x=0; x<m_nIndividualCount; x++ )
	{	
		int nSelection = rand()%list.GetCount();
		POSITION pos = list.FindIndex( nSelection );
		TINDIV* pSelect = ( TINDIV* )list.GetAt( pos );
		pTempPop[ x ] = pSelect->Clone();
	}

	for( x=0; x<m_nIndividualCount; x++  )
	{
		TINDIV* pTemp = m_pIndividuals[ x ];
		delete pTemp;
		m_pIndividuals[ x ] = pTempPop[ x ]; //SetIndividual
		m_pIndividuals[ x ]->SetChecked( false );
	}
	delete[] pTempPop;
}

//////////////////////////////////////////////////////////////////////
// Function Name:	GenitorSelection
// Parameters	:	int nReplaceCount 
// Return Type	:	void
// Description	:	a natural selection method
//					Replace the best ones, with the worst ones..	
// Assumptions	:	Assumed that, all individuals are created explicitly and set.	
//					Assumed that, it is SORTED
///////////////////////////////////
template<class TINDIV, class TBIT >
///////////////////////////////////
void CPopulation<TINDIV, TBIT>::GenitorSelection( int nReplaceCount )
{	
	if( nReplaceCount > m_nIndividualCount / 2 ) //meaningless
	{
		nReplaceCount = m_nIndividualCount / 2;
	}
		
	for( int x=0; x<nReplaceCount; x++ )
	{
		TINDIV* pBest  = m_pIndividuals[ x ]->Clone();
		TINDIV* pWorst = m_pIndividuals[ m_nIndividualCount - x -1 ];
		delete pWorst;
		m_pIndividuals[ m_nIndividualCount-x-1] = pBest;
	}
}

//////////////////////////////////////////////////////////////////////
// Function Name:	TournamentSelection
// Parameters	:	double fAveragePerformance, of the population
// Return Type	:	void
// Description	:	a natural selection method
//					--------------------------
//					Steps:
//					--------------------------
//					-Choose randomly two individual	
//					-toss an UNFAIR coin[ weighted fChanceOfGood, (1-fChanceOfGood)] 
//					-if coin says that "choose good one", choose the individual who has the better performance.
//					-if coin says that "choose bad one", choose the individual who has the worse performance.	
//					--------------------------
// Assumptions	:	Assumed that, all individuals are created explicitly and set.	
///////////////////////////////////
template<class TINDIV, class TBIT >
///////////////////////////////////
void CPopulation<TINDIV, TBIT>::TournamentSelection( double fChanceOfGood )
{
	TINDIV** pNewOnes = new TINDIV*[ m_nIndividualCount ];

	int x=0;
	for( x=0; x<m_nIndividualCount; x++ )
	{
		TINDIV* pFirst	 = GetRandomIndividual();
		TINDIV* pSecond = GetRandomIndividual();
		
		TINDIV* pBetter;
		TINDIV* pWorse;
		if( pFirst->GetPerformance() < pSecond->GetPerformance() )
		{
			pBetter = pSecond;
			pWorse  = pFirst;
		}
		else
		{
			pBetter = pFirst;
			pWorse  = pSecond;
		}
		
		bool bChooseGoodOne = ( ( rand() % RANDOM_INT_PRECISION_ER ) <= ( fChanceOfGood * RANDOM_INT_PRECISION_ER ) );
		if( bChooseGoodOne )
		{
			pNewOnes[ x ] = pBetter->Clone();		
		}
		else
		{
			pNewOnes[ x ] = pWorse->Clone();		
		}
	}
	
	for( x=0; x<m_nIndividualCount; x++ )
	{
		TINDIV* pTemp = m_pIndividuals[ x ];	
		delete pTemp;
		m_pIndividuals[ x ] = pNewOnes[ x ];
	}

	delete[] pNewOnes;
}

//////////////////////////////////////////////////////////////////////
// Function Name:	GetRandomIndividual
// Parameters	:	void
// Return Type	:	TINDIV*
// Description	:	Returns an individual randomly
// Assumptions	:	Assumed that, all individuals are created explicitly and set.	
///////////////////////////////////
template<class TINDIV, class TBIT >
///////////////////////////////////
TINDIV* CPopulation<TINDIV, TBIT>::GetRandomIndividual( )
{
	int nSel = rand() % m_nIndividualCount;
	return m_pIndividuals[ nSel ];	
}

///////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////

#endif // !defined(__POPULATION_H__)
